"""
A file containing classes related to component representation.
There are currently 6 different ways of processing representations. The type of processing is selected based on
structures.xml file in the following way:
if rendition="lemma" --> LemmaCR
elif rendition="lexis" --> LexisCR
elif rendition="word_form" and selection="agreement" --> WordFormAgreementCR
elif rendition="word_form" and selection="msd" --> WordFormMsdCR
elif rendition="word_form" and selection="all" --> WordFormAllCR
elif rendition="word_form" --> WordFormAnyCR

Combining selections has so far been needed only for selection="agreement" and selection="msd", so this was tested,
whereas combinations with selection="all" have so far not been needed, and hence haven't been tested.
"""

import logging
from ast import literal_eval

from collections import Counter

from conversion_utils import jos_msds_and_properties

from cordex.utils.converter import msd_to_properties, default_msd_to_properties
from cordex.words.word import WordDummy


class ComponentRepresentation:
    def __init__(self, data, word_renderer):
        self.data = data
        self.word_renderer = word_renderer

        self.words = []
        self.rendition_text = None
        self.rendition_msd = None

        self.processed_api = False
        self.api_request = []
        self.agreement = []

    def get_agreement_head_component_id(self):
        """ By default, there are no agreements. """
        return []

    def add_word(self, word, is_ud):
        """ Adds word to representation. """
        self.words.append(word)

    def render(self, is_ud, lookup_lexicon=None, lookup_api=None):
        """ Render when text is not already rendered. """
        if self.rendition_text is None:
            self.rendition_text, self.rendition_msd = self._render(is_ud, lookup_lexicon=lookup_lexicon, lookup_api=lookup_api)

        if 'format' in self.data and self.data['format'] == 'lowercase':
            self.rendition_text = self.rendition_text.lower()


    # Convert output to same format as in conllu
    @staticmethod
    def convert_dict_to_string(dictionary):
        if type(dictionary) == str:
            return dictionary
        result = []
        for k, v in dictionary.items():
            result.append(f'{k}={v}')

        return '|'.join(result)

    def _render(self, is_ud, lookup_lexicon=None, lookup_api=None):
        raise NotImplementedError("Not implemented for class: {}".format(type(self)))


class LemmaCR(ComponentRepresentation):
    """ Handles lemma as component representation. """
    def _render(self, is_ud, lookup_lexicon=None, lookup_api=None):

        if len(self.words) > 0:
            lemma = self.words[0].lemma
            if is_ud:
                pos = self.convert_dict_to_string(self.words[0].udpos)
            else:
                pos = self.words[0].xpos

                if lookup_api is not None:
                    properties = default_msd_to_properties(pos, 'en', lemma=lemma)
                    if not lookup_api.executed:
                        self.processed_api = True
                        self.api_request.append({
                            'type': 'lemma',
                            'lemma': lemma,
                            'lemma_features': properties.lexeme_feature_map,
                            "category": properties.category
                        })
                    else:
                        lemma_key = f"{lemma}_{properties.category}_{'_'.join([f'{k},{v}' for k, v in properties.lexeme_feature_map.items()])}"
                        form = lookup_api.find_lemma(lemma_key)
                        if form is not None:
                            pos = form[3]

                elif lookup_lexicon is not None:
                    msd, lemma, text = lookup_lexicon.get_word_form(lemma, pos, self.data, find_lemma_msd=True)
                    if msd is not None:
                        pos = msd
            return self.words[0].lemma, pos
        else:
            return None, None


class LexisCR(ComponentRepresentation):
    """ Handles fixed word as component representation. """
    def _render(self, is_ud, lookup_lexicon=None, lookup_api=None):
        if is_ud:
            pos = 'POS=PART'
        else:
            pos = 'Q'
        return self.data['lexis'], pos


class WordFormAllCR(ComponentRepresentation):
    """ Returns all possible word forms separated with '/' as component representation. """
    def _render(self, is_ud, lookup_lexicon=None, lookup_api=None):
        if len(self.words) == 0:
            return None, None
        else:
            forms = [w.text.lower() for w in self.words]
            if is_ud:
                msds = [self.convert_dict_to_string(w.udpos) for w in self.words]
            else:
                msds = [w.xpos for w in self.words]

            # sorts unique forms by alphabet and finds corresponding msds
            representation_forms = []
            representation_msds = []
            for form in sorted(set(forms)):
                ind = forms.index(form)
                representation_forms.append(forms[ind])
                representation_msds.append(msds[ind])

            return "/".join(representation_forms), "/".join(representation_msds)


class WordFormAnyCR(ComponentRepresentation):
    """ Returns any possible word form as component representation. """
    def _render(self, is_ud, lookup_lexicon=None, lookup_api=None):
        words_counter = []
        for word in self.words:
            if is_ud:
                words_counter.append((str(word.udpos), word.lemma))
            else:
                words_counter.append((word.xpos, word.lemma))
        words_counter_ordered = sorted(list(set(words_counter)),
                                 key=lambda x: (True, x[0], x[1]) if x[1] is not None else (False, '_', '_'))
        sorted_words = sorted(
            words_counter_ordered, key=lambda x: -words_counter.count(x) + (sum(ord(l) for l in x[1]) / 1e5 if x[1] is not None else .5))

        # so lets get through all words, sorted by frequency
        for word_msd, word_lemma in sorted_words:
            # check if agreements match
            agreements_matched = [agr.match(word_msd, is_ud) for agr in self.agreement]

            # if lookup_api exists create queries for it
            if lookup_api is not None:
                self.processed_api = True

                for i, agr in enumerate(self.agreement):
                    if not agr.match(word_msd, is_ud) or not lookup_api.executed:
                        properties = default_msd_to_properties(agr.msd(), 'en', agr.lemma)

                        # get agreement feature properties
                        agreement_properties = default_msd_to_properties(word_msd, 'en', word_lemma)
                        if 'msd' in agr.data:
                            form_features = agr.data['msd']
                        else:
                            form_features = {}

                        missing_form_features = False
                        for agreement_name in agr.data['agreement']:
                            if agreement_name in agreement_properties.form_feature_map:
                                form_features[agreement_name] = agreement_properties.form_feature_map[agreement_name]
                            elif agreement_name in agreement_properties.lexeme_feature_map:
                                form_features[agreement_name] = agreement_properties.lexeme_feature_map[agreement_name]

                            if agreement_name not in agreement_properties.form_feature_map and agreement_name not in agreement_properties.lexeme_feature_map:
                                missing_form_features = True
                                break

                        if missing_form_features:
                            continue

                        if not lookup_api.executed:
                            self.api_request.append({
                                'type': 'agreement',
                                'lemma': agr.lemma,
                                'lemma_features': properties.lexeme_feature_map,
                                'form_features': form_features,
                                "category": properties.category
                            })
                        else:
                            lemma_key = f"{agr.lemma}_{properties.category}_{'_'.join([f'{k},{v}' for k, v in properties.lexeme_feature_map.items()])}"

                            form = lookup_api.find_form(lemma_key, form_features)

                            if form is not None:
                                text = form[0]
                                msd = form[3]
                                agr.msds[0] = msd
                                agr.words.append(WordDummy(msd, agr.lemma, text))
                                # when we find element in sloleks automatically add it (no need for second checks, since msd
                                # is tailored to pass tests by default)
                                agr.rendition_candidate = text
                                agr.rendition_msd_candidate = msd
                                agreements_matched[i] = True
                            else:
                                break

            # in case all agreements do not match try to get data from sloleks and change properly
            elif lookup_lexicon is not None and not all(agreements_matched):
                for i, agr in enumerate(self.agreement):
                    if not agr.match(word_msd, is_ud):
                        msd, lemma, text = lookup_lexicon.get_word_form(agr.lemma, agr.msd(), agr.data, align_msd=word_msd, align_lemma=word_lemma)
                        if msd is not None:
                            agr.msds[0] = msd
                            agr.words.append(WordDummy(msd, lemma, text))
                            # when we find element in sloleks automatically add it (no need for second checks, since msd
                            # is tailored to pass tests by default)
                            agr.rendition_candidate = text
                            agr.rendition_msd_candidate = msd
                            agreements_matched[i] = True
                        else:
                            break

            # if we are at the last "backup word", then confirm matches 
            # that worked for this one and return
            if word_lemma is None:
                for agr, matched in zip(self.agreement, agreements_matched):
                    if matched:
                        agr.confirm_match()
                return None, None

            # if all agreements match, we win!
            if all(agreements_matched):
                for agr in self.agreement:
                    agr.confirm_match()

                if is_ud:
                    word_msd_eval = self.convert_dict_to_string(literal_eval(word_msd))
                else:
                    word_msd_eval = word_msd

                text = self.word_renderer.render(word_lemma, word_msd)

                if text:
                    return self.word_renderer.render(word_lemma, word_msd), word_msd_eval
                else:
                    if len(self.words) != 1:
                        raise ValueError('Internal database is not setup correctly! All msd - lemma combinations should be in UniqWords table. In this situation they are not.')
                    else:
                        return self.words[0].text, word_msd_eval
        return None, None


class WordFormMsdCR(WordFormAnyCR):
    """ Handles component representations with msd requirement. If none of the words in input satisfy checks (ie. no
    word with such msd), try to form word from database lookups. """
    def __init__(self, *args):
        super().__init__(*args)
        self.lemma = None
        self.msds = []
    
    def msd(self):
        return self.msds[0]

    def check_udpos(self, word_msd):
        """ Checks whether word msd matches the one specified in structures file. """
        if 'msd' not in self.data:
            return True
        selectors = self.data['msd']

        for key, value in selectors.items():
            if key not in word_msd or word_msd[key] != value:
                return False

        return True

    def check_xpos(self, word_msd, word_lemma):
        """ Checks whether word msd matches the one specified in structures file. """
        if 'msd' not in self.data:
            return True
        selectors = self.data['msd']

        properties = msd_to_properties(word_msd, 'en', lemma=word_lemma)
        for key, value in selectors.items():
            key_lower = key.lower()
            if key_lower in properties and properties[key_lower] != value:
                return False

        return True

    def add_word(self, word, is_ud):
        """ Adds lemma and msd. Also adds word if word matches msd check. """
        if self.lemma is None:
            self.lemma = word.lemma

        if is_ud:
            self.msds.append(word.udpos)
            if self.check_udpos(word.udpos):
                super().add_word(word, is_ud)
        else:
            self.msds.append(word.xpos)
            if self.check_xpos(word.xpos, word.lemma):
                super().add_word(word, is_ud)

    def _render(self, is_ud, lookup_lexicon=None, lookup_api=None):
        if is_ud:
            self.words.append(WordDummy(self._common_udpos()))
        else:
            word_appended = False
            if len(self.words) == 0 and lookup_api is not None:
                properties = default_msd_to_properties(self.msd(), 'en', self.lemma)
                if not lookup_api.executed:
                    self.api_request.append({
                        'type': 'msd',
                        'lemma': self.lemma,
                        'lemma_features': properties.lexeme_feature_map,
                        'form_features': self.data['msd'],
                        "category": properties.category
                    })
                    self.processed_api = True
                    self.words.append(WordDummy(self.msd(), self.lemma, ''))
                    word_appended = True
                else:
                    lemma_key = f"{self.lemma}_{properties.category}_{'_'.join([f'{k},{v}' for k, v in properties.lexeme_feature_map.items()])}"
                    lemma = self.lemma
                    form = lookup_api.find_form(lemma_key, self.data['msd'])
                    if form is not None:
                        text = form[0]
                        msd = form[3]
                        self.words.append(WordDummy(msd, lemma, text))
                        word_appended = True

            elif len(self.words) == 0 and lookup_lexicon is not None:
                msd, lemma, text = lookup_lexicon.get_word_form(self.lemma, self.msd(), self.data)
                if msd is not None:
                    self.words.append(WordDummy(msd, lemma, text))
                    word_appended = True
            if not word_appended:
                self.words.append(WordDummy(self._common_xpos()))
        return super()._render(is_ud, lookup_lexicon, lookup_api)
    
    def _common_xpos(self):
        """ Tries to form xpos that is present in all examples. """
        msds = sorted(self.msds, key=len)
        common_msd = ["-" if not all(msds[j][idx] == msds[0][idx] for j in range(1, len(self.msds))) 
                      else msds[0][idx] for idx in range(len(msds[0]))]
        common_msd = "".join(common_msd)
        return common_msd

    def _common_udpos(self):
        """ Tries to form udpos that is present in all examples. """
        udpos_key = self.msds[0]
        udpos_key_new = set(self.msds[0].keys())
        for msd in self.msds:
            for k, v in udpos_key.items():
                if k not in msd or msd[k] != v:
                    udpos_key_new.remove(k)

            udpos_key = {k: v for k, v in udpos_key.items() if k in udpos_key_new}
        return udpos_key


class WordFormAgreementCR(WordFormMsdCR):
    """ Handles word forms with agreement. """
    def __init__(self, data, word_renderer):
        super().__init__(data, word_renderer)
        self.rendition_candidate = None
        self.rendition_msd_candidate = None

    def get_agreement_head_component_id(self):
        """ Returns agreement head component id. """
        return self.data['other']

    def match(self, word_msd, is_ud):
        """ Checks if word_msd matches any of possible words in agreements. """
        lemma_available_words = self.word_renderer.available_words(self.lemma)
        for candidate_pos, candidate_text, candidate_lemma in lemma_available_words:
            if is_ud:
                if self.msd()['POS'] != candidate_pos['POS']:
                    continue

                if WordFormAgreementCR.check_agreement_udpos(word_msd, candidate_pos, self.data['agreement']):
                    if self.check_udpos(candidate_pos):
                        self.rendition_candidate = candidate_text
                        self.rendition_msd_candidate = self.convert_dict_to_string(candidate_pos)
                        return True
            else:
                if self.msd()[0] != candidate_pos[0]:
                    continue

                if WordFormAgreementCR.check_agreement(word_msd, candidate_pos, self.data['agreement']):
                    if self.check_xpos(candidate_pos, candidate_lemma):
                        self.rendition_candidate = candidate_text
                        self.rendition_msd_candidate = candidate_pos
                        return True

        return False

    def confirm_match(self):
        """ Stores final state.  """
        self.rendition_text = self.rendition_candidate
        if 'format' in self.data and self.data['format'] == 'lowercase':
            self.rendition_text = self.rendition_text.lower()
        self.rendition_msd = self.rendition_msd_candidate

    @staticmethod
    def check_agreement_udpos(msd1, msd2, agreements):
        """ Checks if msds match in agreements. """
        msd1 = literal_eval(msd1)
        for agr_case in agreements:
            agr_case = agr_case.capitalize()

            # if not in msd, some strange msd was tries, skipping...
            if agr_case not in msd1:
                logging.warning("Cannot do agreement: {} for msd {} not found!"
                                .format(agr_case, msd1))
                return False

            if agr_case not in msd2:
                logging.warning("Cannot do agreement: {} for msd {} not found!"
                                .format(agr_case, msd2))
                return False

            # match!
            if msd1[agr_case] != msd2[agr_case]:
                return False

        return True

    @staticmethod
    def check_agreement(msd1, msd2, agreements):
        """ Checks if msds match in agreements. """
        properties1 = msd_to_properties(msd1, 'en')
        properties2 = msd_to_properties(msd2, 'en')
        for agr_case in agreements:
            # if not in msd, some strange msd was tries, skipping...
            if agr_case not in properties1:
                return False

            if agr_case not in properties2:
                return False

            p1 = properties1[agr_case]
            p2 = properties2[agr_case]

            # match!
            if p1 != p2:
                return False

        return True

    def render(self, is_ud, lookup_lexicon=None, lookup_api=None):
        pass



CODES_UD = {
    "ADJ",
    "ADP",
    "PUNCT",
    "ADV",
    "AUX",
    "SYM",
    "INTJ",
    "CCONJ",
    "X",
    "NOUN",
    "DET",
    "PROPN",
    "NUM",
    "VERB",
    "PART",
    "PRON",
    "SCONJ"
}

CODES = {
    "Noun": "N",
    "Verb": "V",
    "Adjective": "A",
    "Adverb": "R",
    "Pronoun": "P",
    "Numeral": "M",
    "Preposition": "S",
    "Conjunction": "C",
    "Particle": "Q",
    "Interjection": "I",
    "Abbreviation": "Y",
    "Residual": "X",
    "Punctuation": "Z",

    'common': 'c',
    'proper': 'p',
    'masculine': 'm',
    'feminine': 'f',
    'neuter': 'n',
    "singular": "s",
    "dual": "d",
    "plural": "p",
    "nominative": "n",
    "genitive": "g",
    "dative": "d",
    "accusative": "a",
    "locative": "l",
    "instrumental": "i",
    "no": "n",
    "yes": "y",
    "main": "m",
    "auxiliary": "a",
    "perfective": "e",
    "progressive": "p",
    "biaspectual": "b",
    "infinitive": "n",
    "supine": "u",
    "participle": "p",
    "present": "r",
    "future": "f",
    "conditional": "c",
    "imperative": "m",
    "first": "1",
    "second": "2",
    "third": "3",
    "general": "g",
    "possessive": "s",
    "positive": "p",
    "comparative": "c",
    "superlative": "s",
    "personal": "p",
    "demonstrative": "d",
    "relative": "r",
    "reflexive": "x",
    "interrogative": "q",
    "indefinite": "i",
    "negative": "z",
    "bound": "b",
    "digit": "d",
    "roman": "r",
    "letter": "l",
    "cardinal": "c",
    "ordinal": "o",
    "pronominal": "p",
    "special": "s",
    "coordinating": "c",
    "subordinating": "s",
    "foreign": "f",
    "typo": "t",
    "program": "p",
    "web": "w",
    "emo": "e",
    "hashtag": "h",
    "at": "a"
}

TAGSET = {
    "N": ['type', 'gender', 'number', 'case', 'animate'],
    "V": ['type', 'aspect', 'vform', 'person', 'number', 'gender', 'negative'],
    "A": ['type', 'degree', 'gender', 'number', 'case', 'definiteness'],
    "R": ['type', 'degree'],
    "P": ['type', 'person', 'gender', 'number', 'case', 'owner_number', 'owned_gender', 'clitic'],
    "M": ['form', 'type', 'gender', 'number', 'case', 'definiteness'],
    "S": ['case'],
    "C": ['type'],
    "Q": [],
    "I": [],
    "Y": [],
    "X": ['type']
}

PPB_DEPRELS = [
    "advmod",
    "amod",
    "compound",
    "conj",
    "fixed",
    "flat",
    "iobj",
    "nmod",
    "nsubj",
    "nummod",
    "obj",
    "obl"
]



"""
Class for grouping restrictions.
"""
from cordex.restrictions.restriction import Restriction


class RestrictionGroup:
    def __init__(self, restrictions_tag, is_ud, group_type='and'):
        self.restrictions = [Restriction(el, is_ud) for el in restrictions_tag]
        self.group_type = group_type

    def __iter__(self):
        for restriction in self.restrictions:
            yield restriction

    def match(self, word):
        """ Checks whether sufficient restrictions are met. """
        if self.group_type == 'or':
            for restr in self.restrictions:
                if restr.match(word):  # match either
                    return True
            return False
        elif self.group_type == 'and':
            for restr in self.restrictions:
                if not restr.match(word):  # match and
                    return False
            return True
        else:
            raise Exception("Unsupported group_type - it may only be 'and' or 'or'")



"""
Classes for handling restrictions.
"""
from enum import Enum
from cordex.utils.converter import msd_to_properties


class RestrictionType(Enum):
    Morphology = 0
    Lexis = 1
    MatchAll = 2
    Space = 3
    MorphologyUD = 4


def determine_ppb_ud(rgxs):
    """ Determine if a word has full meaning - UD (returns priority). """
    if len(rgxs) != 1:
        return 0
    rgx = rgxs[0]
    if rgx in ("ADJ", "NOUN", "ADV"):
        return 0
    elif rgx == "AUX":
        return 3
    elif rgx == "VERB":
        return 2
    else:
        return 4


def determine_ppb(restrictions):
    """ Determine if a word has full meaning - JOS (returns priority). """
    if len(restrictions['POS'][0]) != 1:
        return 0
    category = list(restrictions['POS'][0])[0]
    # rgx = rgxs[0]
    if category in ("adjective", "noun", "adverb"):
        return 0
    elif category == "verb":
        if len(restrictions) == 1:
            return 2
        type = list(restrictions['type'][0])[0] if 'type' in restrictions else ''
        if type == 'auxiliary':
            return 3
        elif type == 'main':
            return 1
        else:
            return 2
    else:
        return 4


class MorphologyRegex:
    """
    Class in charge of morphology JOS restriction regexes.
    """
    def __init__(self, restriction):
        restr_dict = {}
        for feature in restriction:
            feature_dict = dict(feature.items())

            negate = False
            if "filter" in feature_dict:
                assert feature_dict['filter'] == "negative"
                negate = True
                del feature_dict['filter']

            assert len(feature_dict) == 1
            key, value = next(iter(feature_dict.items()))
            restr_dict[key] = (set(value.split('|')), negate)

        assert 'POS' in restr_dict
        self.restrictions = restr_dict

    def __call__(self, text, lemma):
        if not text:
            return False

        properties = msd_to_properties(text, 'en', lemma)

        for res_name, res_val in self.restrictions.items():
            res_name = res_name.lower()

            # handles category
            if res_name == 'pos':
                if properties['pos'] not in res_val[0]:
                    return False

            # handles form other restrictions
            # handles restrictions where we negate filter is off
            elif res_val[1] == False:
                if res_name in properties:
                    if not properties[res_name] in res_val[0]:
                        return False
                else:
                    return False
            # handle restrictions with negate filter on
            elif res_val[1] == True:
                if res_name in properties:
                    if properties[res_name] in res_val[0]:
                        return False

        return True


class MorphologyUDRegex:
    """
    Class in charge of morphology UD restriction regexes.
    """
    def __init__(self, restriction):
        restr_dict = {}
        for feature in restriction:
            feature_dict = dict(feature.items())

            negate = False
            if "filter" in feature_dict:
                negate = True
                del feature_dict['filter']

            assert len(feature_dict) == 1
            key, value = next(iter(feature_dict.items()))
            restr_dict[key] = (set(value.split('|')), negate)

        assert 'POS' in restr_dict

        self.restrictions = restr_dict

    def __call__(self, msd, lemma):
        udpos = msd.udpos
        if not udpos:
            return False

        for res_name, res_val in self.restrictions.items():
            # handles form other restrictions
            # handles restrictions where we negate filter is off
            if res_val[1] == False:
                if res_name in udpos:
                    if not udpos[res_name] in res_val[0]:
                        return False
                else:
                    return False
            # handle restrictions with negate filter on
            elif res_val[1] == True:
                if res_name in udpos:
                    if udpos[res_name] in res_val[0]:
                        return False

        return True


class LexisRegex:
    """
    Class in charge of checking whether lemma is in specified list of words.
    """
    def __init__(self, restriction):
        restr_dict = {}
        for feature in restriction:
            restr_dict.update(feature.items())

        assert "lemma" in restr_dict
        self.match_list = restr_dict['lemma'].split('|')
    
    def __call__(self, text, lemma):
        return text in self.match_list


class SpaceRegex:
    def __init__(self, restriction):
        restr_dict = {}
        for feature in restriction:
            restr_dict.update(feature.items())

        assert "contact" in restr_dict
        self.space = restr_dict['contact'].split('|')
        for el in self.space:
            if el not in ['both', 'right', 'left', 'neither']:
                raise Exception('Value of space restriction is not supported (it may be both, left, right or neither).')

    def __call__(self, word, lemma):
        match = False
        if 'neither' in self.space:
            match = match or (not word.previous_glue and not word.glue)
        if 'left' in self.space:
            match = match or (word.previous_glue and not word.glue)
        if 'right' in self.space:
            match = match or (not word.previous_glue and word.glue)
        if 'both' in self.space:
            match = match or (word.previous_glue and word.glue)

        return match


class Restriction:
    """ A class containing restriction. """
    def __init__(self, restriction_tag, is_ud):
        self.ppb = 4  # polnopomenska beseda (0-4)

        if restriction_tag is None:
            self.type = RestrictionType.MatchAll
            self.matcher = None
            self.present = None
            return

        restriction_type = restriction_tag.get('type')
        if restriction_type == "morphology":
            # UD system is handled based on deprel
            if is_ud:
                self.type = RestrictionType.MorphologyUD
                self.matcher = MorphologyUDRegex(list(restriction_tag))
            #     self.ppb = determine_ppb_ud(self.matcher.rgxs)
            else:
                self.type = RestrictionType.Morphology
                self.matcher = MorphologyRegex(list(restriction_tag))
                self.ppb = determine_ppb(self.matcher.restrictions)

        elif restriction_type == "lexis":
            self.type = RestrictionType.Lexis
            self.matcher = LexisRegex(list(restriction_tag))

        elif restriction_type == "space":
            self.type = RestrictionType.Space
            self.matcher = SpaceRegex(list(restriction_tag))
        else:
            raise NotImplementedError()

    def match(self, word):
        """ Obtains data necessary for matcher and runs it. """
        if self.type == RestrictionType.Morphology:
            match_to = word.xpos
        elif self.type == RestrictionType.MorphologyUD:
            match_to = word
        elif self.type == RestrictionType.Lexis:
            match_to = word.lemma
        elif self.type == RestrictionType.MatchAll:
            return True
        elif self.type == RestrictionType.Space:
            match_to = word
        else:
            raise RuntimeError("Unreachable!")

        return self.matcher(match_to, word.lemma)



"""
Custom processing after matching.
"""


class Postprocessor:
    def __init__(self, fixed_restriction_order=False, lang='sl'):
        self.fixed_restriction_order = fixed_restriction_order
        self.lang = lang

    @staticmethod
    def fix_sz_sl(next_word):
        """ Fixes s/z in Slovenian. """
        if next_word[0] in ['c', 'č', 'f', 'h', 'k', 'p', 's', 'š', 't']:
            return 's'
        return 'z'

    @staticmethod
    def fix_kh_sl(next_word):
        """ Fixes k/h in Slovenian. """
        if next_word[0] in ['g', 'k']:
            return 'h'
        return 'k'

    def process(self, match, collocation_id):
        """ Runs Slovenian k/h and s/z fixes. """
        if self.lang != 'sl':
            return match, collocation_id

        if len(collocation_id) > 2:
            for idx, (col_id, word) in enumerate(collocation_id[1:-1]):
                if word in ['s', 'z']:
                    correct_letter = self.fix_sz_sl(collocation_id[idx + 2][1])
                    collocation_id[idx + 1][1] = correct_letter
                    match[col_id].text = correct_letter
                elif word in ['k', 'h']:
                    correct_letter = self.fix_kh_sl(collocation_id[idx + 2][1])
                    collocation_id[idx + 1][1] = correct_letter
                    match[col_id].text = correct_letter
        collocation_id = [collocation_id[0]] + [tuple(line) for line in collocation_id[1:]]
        return match, collocation_id

    def is_fixed_restriction_order(self, match):
        """ If necessary, checks if restrictions are in correct order. """

        if not self.fixed_restriction_order:
            return True

        sorted_dict = {k: v for k, v in sorted(match.items(), key=lambda item: item[1].int_id)}
        prev_id = -1
        for key in sorted_dict.keys():
            if key == '#':
                continue
            int_key = int(key)
            if prev_id > int_key:
                return False
            prev_id = int_key

        return True



"""
A file in charge of properly formatting output text.
"""
from math import log2
import re
import logging

from cordex.structures.component import ComponentType


class Formatter:
    def __init__(self, collocation_ids, word_renderer, is_ud, args):
        self.collocation_ids = collocation_ids
        self.word_renderer = word_renderer
        self.is_ud = is_ud
        self.decimal_separator = args['decimal_separator']
        self.args = args
        self.additional_init()
    
    def header_repeat(self):
        raise NotImplementedError("Header repeat formatter not implemented")

    def header_right(self):
        raise NotImplementedError("Header right formatter not implemented")

    def content_repeat(self, words, representations, idx, sidx):
        raise NotImplementedError("Content repeat formatter not implemented")

    def content_right(self, freq):
        raise NotImplementedError("Content right formatter not implemented")
    
    def additional_init(self):
        pass

    def length(self):
        return len(self.header_repeat())

    def set_structure(self, structure):
        pass

    def new_match(self, match):
        pass


class OutNoStatFormatter(Formatter):
    """
    Formats regular output text without statistics.
    """
    def additional_init(self):
        self.representation = {}

    def header_repeat(self):
        """ Returns no stats header that occurs multiple times. """
        return ["Lemma", "Representative_form", "RF_msd", "RF_scenario"]
    
    def header_right(self):
        """ Returns no stats header for columns that occur one time. """
        return ["Joint_representative_form_fixed", "Joint_representative_form_variable", "Frequency"]
    
    def content_repeat(self, words, representations, idx, _sidx):
        """ Returns no stats content for columns that might appear multiple times. """
        word = words[idx]
        if idx not in representations:
            return ["", "", ""]

        rep_text, rep_msd = representations[idx]
        if rep_text is None:
            self.representation[idx] = word.lemma
            return [word.lemma, word.lemma, "", "lemma_fallback"]
        else:
            self.representation[idx] = rep_text
            return [word.lemma, rep_text, rep_msd, "ok"]

    def content_right(self, freq, variable_word_order=None):
        """ Returns no stats content for columns that occur one time. """
        fixed_word_order = sorted(self.representation.keys())
        if variable_word_order is None:
            variable_word_order = fixed_word_order
        rep_fixed_word_order = ' '.join([self.representation[o] for o in fixed_word_order if o in self.representation])
        rep_variable_word_order = ' '.join([self.representation[o] for o in variable_word_order if o in self.representation])
        result = [rep_fixed_word_order, rep_variable_word_order, str(freq)]
        self.representation = {}
        return result
    
    def __str__(self):
        return "out-no-stat"


class StatsFormatter(Formatter):
    """
    Formats text for statistics.
    """
    def additional_init(self):
        self.stats = None
        self.jppb = None
        self.corew = None
    
    @staticmethod
    def stat_str(num, decimal_separator):
        string = "{:.5f}".format(num) if isinstance(num, float) else str(num)
        if decimal_separator != '.':
            string = string.replace('.', decimal_separator)
        return string
    
    def set_structure(self, structure):
        """ Gets full meaning words positions from structure. """
        jppb = []
        corew = []

        for component in structure.components:
            if component.type == ComponentType.Core2w:
                jppb.append(component.idx)
            if component.type != ComponentType.Other:
                corew.append(component.idx)

        assert(len(jppb) == 2)
        self.jppb = tuple(jppb)
        self.corew = tuple(corew)
    
    def new_match(self, match):
        """ Calculates statistics for a match. """
        self.stats = {"freq": {}}

        for cid in self.corew:
            if cid not in match.matches[0]:
                freq = 0
            else:
                word = match.matches[0][cid]
                if self.is_ud:
                    word_msd0 = word.udpos['POS']
                else:
                    word_msd0 = word.xpos[0]

                freq = self.word_renderer.num_words(word.lemma, word_msd0)

            self.stats["freq"][cid] = freq

        fx = self.stats["freq"][self.jppb[0]]
        fy = self.stats["freq"][self.jppb[1]]
        freq = len(match)
        N = self.word_renderer.num_all_words()

        self.stats['d12'] = freq / fx - (fy - freq) / (N - fx)
        self.stats['d21'] = freq / fy - (fx - freq) / (N - fy)

        self.stats['df'] = match.distinct_forms()
        self.stats['freq_all'] = freq

    def header_repeat(self):
        """ Returns stats header that occurs multiple times. """
        return ["Distribution"]
    
    def header_right(self):
        """ Returns stats header that occur one time only. """
        return ["Delta_p12", "Delta_p21", "LogDice_core", "LogDice_all", "Distinct_forms"]
    
    def content_repeat(self, words, representations, idx, sidx):
        """ Returns stats for columns that might appear multiple times. """
        # not a core word
        if idx not in self.corew:
            return [""] * self.length()

        word = words[idx]
        key = (sidx, idx, word.lemma)
        # try to fix missing dispersions
        if key not in self.collocation_ids.dispersions:
            if word.lemma == 'k':
                new_key = (sidx, idx, 'h')
            elif word.lemma == 'h':
                new_key = (sidx, idx, 'k')
            elif word.lemma == 's':
                new_key = (sidx, idx, 'z')
            elif word.lemma == 'z':
                new_key = (sidx, idx, 's')
            else:
                new_key = (sidx, idx, '')
            if new_key in self.collocation_ids.dispersions:
                key = new_key
                logging.info('Dispersions fixed.')
            else:
                logging.info('Dispersions not fixed.')
        if key in self.collocation_ids.dispersions:
            distribution = self.collocation_ids.dispersions[key]
        else:
            distribution = 1
        return [self.stat_str(distribution, self.decimal_separator)]
    
    def content_right(self, freq):
        """ Returns stats for columns that occur one time only. """
        fx = self.stats["freq"][self.jppb[0]]
        fy = self.stats["freq"][self.jppb[1]]
        freq = self.stats['freq_all']
        logdice_core = 14 + log2(2 * freq / (fx + fy))

        fi = [self.stats["freq"][idx] for idx in self.corew]
        fi = [f for f in fi if f > 0]
        logdice_all = 14 + log2(len(fi) * freq / sum(fi))

        return [self.stat_str(x, self.decimal_separator) for x in (
            self.stats["d12"], self.stats["d21"], logdice_core, logdice_all, self.stats['df']
        )]
    
    def __str__(self):
        return "stat"

class OutFormatter(Formatter):
    """
    Formats regular output text.
    """
    def additional_init(self):
        self.f1 = OutNoStatFormatter(self.collocation_ids, self.word_renderer, self.is_ud, self.args)
        self.f2 = StatsFormatter(self.collocation_ids, self.word_renderer, self.is_ud, self.args)

    def header_repeat(self):
        """ Combines and returns no stats header with stats header on columns that might appear multiple times. """
        return self.f1.header_repeat() + self.f2.header_repeat()

    def header_right(self):
        """ Combines and returns no stats header with stats header columns that occur one time. """
        return self.f1.header_right() + self.f2.header_right()

    def content_repeat(self, words, representations, idx, sidx, variable_word_order=None):
        """ Combines no stats content with stats content on columns that might appear multiple times. """
        cr1 = self.f1.content_repeat(words, representations, idx, sidx)
        cr2 = self.f2.content_repeat(words, representations, idx, sidx)
        return cr1 + cr2

    def content_right(self, freq, variable_word_order=None):
        """ Combines no stats content with stats content on columns that appear only once. """
        return self.f1.content_right(freq, variable_word_order) + self.f2.content_right(freq)

    def set_structure(self, structure):
        """ Sets structure where necessary. """
        self.f2.set_structure(structure)

    def new_match(self, match):
        """ Formats output from a match. """
        self.f2.new_match(match)
    
    def __str__(self):
        return "out"



"""
Corpus loaders.
"""
import os
from xml.etree import ElementTree
import logging
import re
import sys
import pathlib
from io import StringIO
import conllu
from conversion_utils.jos_msds_and_properties import Converter

from cordex.utils.converter import translate_jos_depparse
from cordex.utils.progress_bar import progress
from cordex.words.word import WordUD, WordJOS


def load_files(args, database):
    """ Loads corpora in various formats. """
    filenames = args['corpus']

    if len(filenames) == 1 and os.path.isdir(filenames[0]):
        all_filenames = []
        for path, subdirs, files in os.walk(filenames[0]):
            for name in files:
                if name[-5:] != '.zstd':
                    all_filenames.append(os.path.join(path, name))
        filenames = all_filenames

    if len(filenames) > 1:
        filenames = [filename for filename in filenames if filename[-5:] != '.zstd']
        filenames = sorted(filenames)

    database.init("CREATE TABLE Files ( filename varchar(2048) )")

    for idx, fname in enumerate(filenames):
        logging.info("FILE " + fname + "{}/{}".format(idx, len(filenames)))
        extension = pathlib.Path(fname).suffix

        # check if file with the same name already loaded...
        loaded = database.execute("SELECT * FROM Files WHERE filename=?", (fname,)).fetchone()
        if loaded is not None:
            logging.info("ALREADY LOADED")
            continue

        if extension == ".xml":
            et = load_tei(fname)
            yield tei_sentence_generator(et, args)
        elif extension == ".conllu" or extension == ".conllup":
            yield load_conllu(fname, args)
        else:
            raise Exception(f'File {fname} is in incorrect format (it should be .xml, .conllu or .conllup).')

        database.execute("INSERT INTO Files (filename) VALUES (?)", (fname,))
        database.commit()


def load_conllu(filename, args):
    """ Loads corpus file in conllu format. """
    if args['jos_msd_lang'] == 'sl':
        raise NotImplementedError('jos_msd_lang == "sl" is not implemented for conllu data!')
    result = []

    words = {}
    links = []

    def sentence_end(bad_sentence, sent_id):
        if bad_sentence:
            return

        for lfrom, ldest, ana in links:
            if lfrom not in words or ldest not in words:
                logging.warning("Bad link in sentence: " + sent_id)
                continue
            words[lfrom].add_link(ana, words[ldest])
        result.extend(words.values())

    with open(filename, 'r', encoding="UTF-8") as f:
        data = f.read()

        conlls = conllu.parse_incr(StringIO(data))
        # build dep parse
        for sent in conlls:
            try:
                if len(sent) == 0:
                    continue

                # adding fake word
                if args['is_ud']:
                    words['0'] = WordUD.fake_root_word('0')
                else:
                    words['0'] = WordJOS.fake_root_word('0')

                for word in sent:
                    if type(word['id']) == tuple:
                        continue

                    if args['is_ud']:
                        words[str(word['id'])] = WordUD.from_conllu_element(word, sent)
                        links.append((str(word['head']), str(word['id']), word['deprel']))
                    else:
                        words[str(word['id'])] = WordJOS.from_conllu_element(word, sent)
                        links.append((str(word['head']), str(word['id']), translate_jos_depparse(word['deprel'], args['jos_depparse_lang'] != 'sl')))
                if 'sent_id' in sent.metadata:
                    metadata = sent.metadata['sent_id']
                else:
                    logging.warning(f'At least one sentence is missing `sent_id`. This may lead to incomplete output in collocation_sentence_mapper.')
                    metadata = ''
                sentence_end(False, metadata)
                links = []
                words = {}
            except:
                links = []
                words = {}
                logging.error(f"Error while reading file {filename} in sentence {sent.metadata['sent_id']}. Check if required data is available!")

    return result


def load_tei(filename):
    """ Loads corpus file in TEI format. """
    with open(filename, 'r') as fp:
        content = fp.read()

    xmlstring = re.sub(' xmlns="[^"]+"', '', content, count=1)
    xmlstring = xmlstring.replace(' xml:', ' ')
    return ElementTree.XML(xmlstring)


def tei_sentence_generator(et, args):
    """ Generates sentences from TEI format. """
    do_msd_translate = not args['jos_msd_lang'] == 'en'
    do_msd_translate = Converter() if do_msd_translate else False

    words = {}
    teis = list(et.iter('{http://www.tei-c.org/ns/1.0}TEI'))

    ns = '{http://www.tei-c.org/ns/1.0}' if len(teis) != 0 else ''

    # if not TEI element search for paragraphs only
    teis = [et] if len(teis) == 0 else teis

    for tei in teis:
        paragraphs = list(tei.iter(ns + 'p'))
        for paragraph in progress(paragraphs, "load-text"):
            sentences = list(paragraph.iter(ns + 's'))
            for sentence in sentences:
                # create fake root word
                if args['is_ud']:
                    word = WordUD.fake_root_word(sentence.get('id'))
                else:
                    word = WordJOS.fake_root_word(sentence.get('id'))
                words[sentence.get('id')] = word
                previous_word = word

                for w in sentence.iter():
                    if w.tag == ns + 'w':
                        if args['is_ud']:
                            word = WordUD.from_tei_element(w, do_msd_translate)
                        else:
                            word = WordJOS.from_tei_element(w, do_msd_translate)

                        word.previous_glue = previous_word.glue
                        words[w.get('id')] = word
                        previous_word = word

                    elif w.tag == ns + 'pc':
                        if args['is_ud']:
                            word = WordUD.pc_word(w, do_msd_translate)
                        else:
                            word = WordJOS.pc_word(w, do_msd_translate)

                        word.previous_glue = previous_word.glue
                        words[w.get('id')] = word
                        previous_word = word

                for l in sentence.iter(ns + "link"):
                    if 'dep' in l.keys():
                        ana = l.get('afun')
                        lfrom = l.get('from')
                        dest = l.get('dep')
                    else:
                        ana = l.get('ana')
                        if args['is_ud']:
                            if ana[:7] != 'ud-syn:':  # dont bother...
                                continue
                            ana = ana[7:]
                            lfrom, dest = l.get('target').replace('#', '').split()
                        else:
                            if ana[:8] != 'jos-syn:':  # dont bother...
                                continue
                            ana = ana[8:]
                            lfrom, dest = l.get('target').replace('#', '').split()

                    if lfrom in words:
                        if dest in words:
                            next_word = words[dest]
                            translated_ana = translate_jos_depparse(ana, args['jos_depparse_lang'] != 'sl')
                            words[lfrom].add_link(translated_ana, next_word)
                        else:
                            logging.error("Unknown id: {}".format(dest))
                            sys.exit(1)

                    else:
                        # strange errors, just skip...
                        pass

    return list(words.values())



import os
import requests
from tqdm import tqdm
from pathlib import Path
import json
import hashlib
import zipfile
import shutil
import logging

from cordex._version import __resources_version__

logger = logging.getLogger('cordex')

# set home dir for default
HOME_DIR = str(Path.home())
CORDEX_RESOURCES_GITHUB = 'https://raw.githubusercontent.com/clarinsi/cordex-resources/'
DEFAULT_RESOURCES_URL = os.getenv('CORDEX_RESOURCES_URL', CORDEX_RESOURCES_GITHUB + 'main')
DEFAULT_RESOURCES_VERSION = os.getenv(
    'CORDEX_RESOURCES_VERSION',
    __resources_version__
)
DEFAULT_MODEL_URL = os.getenv('CORDEX_MODEL_URL', 'default')
DEFAULT_MODEL_DIR = os.getenv(
    'CORDEX_RESOURCES_DIR',
    os.path.join(HOME_DIR, 'cordex_resources')
)


def ensure_dir(dir):
    """
    Create dir in case it does not exist.
    """
    Path(dir).mkdir(parents=True, exist_ok=True)


def get_md5(path):
    """
    Get the MD5 value of a path.
    """
    data = open(path, 'rb').read()
    return hashlib.md5(data).hexdigest()


def file_exists(path, md5):
    """
    Check if the file at `path` exists and match the provided md5 value.
    """
    # if os.path.exists(path):
    #     print(f'PATH: {path} || Written md5: {md5} || Calculated md5: {get_md5(path)}')
    return os.path.exists(path) and get_md5(path) == md5


def unzip_file(filename, zipped_filename):
    """
    Unzip only one file and name it as `filename`
    """
    logger.debug(f'Unzip: {zipped_filename} to {filename} ...')
    with zipfile.ZipFile(zipped_filename) as rf:
        files = zipfile.ZipFile.infolist(rf)
        for file in files:
            with open(filename, 'wb') as wf:
                wf.write(rf.read(file.filename))


def download_file(url, path):
    """
    Download a URL into a file as specified by `path`.
    """
    verbose = logger.level in [0, 10, 20]

    # Try to get zipped file
    zipped = True
    r = requests.get(url + '.zip', stream=True)

    # Zipped file not available
    if not r.ok:
        zipped = False
        r = requests.get(url, stream=True)

    if zipped:
        normal_path = path
        path += '.zip'
    with open(path, 'wb') as f:
        file_size = int(r.headers.get('content-length'))
        default_chunk_size = 131072
        desc = 'Downloading ' + url
        with tqdm(total=file_size, unit='B', unit_scale=True, \
            disable=not verbose, desc=desc) as pbar:
            for chunk in r.iter_content(chunk_size=default_chunk_size):
                if chunk:
                    f.write(chunk)
                    f.flush()
                    pbar.update(len(chunk))

    if zipped:
        unzip_file(normal_path, path)
        os.remove(path)


def request_file(url, path, md5=None):
    """
    A complete wrapper over download_file() that also make sure the directory of
    `path` exists, and that a file matching the md5 value does not exist.
    """
    ensure_dir(Path(path).parent)
    if file_exists(path, md5):
        logger.info(f'File exists: {path}.')
        return
    download_file(url, path)
    assert(not md5 or file_exists(path, md5))


def download(
        lang='sl',
        dir=DEFAULT_MODEL_DIR,
        resources_url=DEFAULT_RESOURCES_URL,
        resources_branch=None,
        resources_version=DEFAULT_RESOURCES_VERSION
    ):


    if resources_url == DEFAULT_RESOURCES_URL and resources_branch is not None:
        resources_url = CORDEX_RESOURCES_GITHUB + resources_branch
    # Download resources.json to obtain latest packages.
    logger.debug('Downloading resource file...')
    # make request
    request_file(
        f'{resources_url}/resources_{resources_version}.json',
        os.path.join(dir, 'resources.json')
    )
    # unpack results
    try:
        resources = json.load(open(os.path.join(dir, 'resources.json')))
    except:
        raise Exception(
            f'Cannot load resource file. Please check your network connection, '
            f'or provided resource url and resource version.'
        )
    if lang not in resources:
        raise Exception(f'Unsupported language: {lang}.')

    url = resources['url']

    try:
        request_file(
            f'{url}/{resources[lang]["link"]}',
            os.path.join(dir, f'{lang}.xz'),
            md5=resources[lang]['md5']
        )
    except KeyError as e:
        raise Exception(
            f'Cannot download file at {url}/{resources[lang]["link"]}. Please recheck url.'
        ) from e
    logger.info(f'Finished downloading models and saved to {dir}.')



"""
A file for forming and saving outputs.
"""
import logging
import os
import shutil
from pathlib import Path

from cordex.utils.progress_bar import progress
from cordex.writers.formatter import OutFormatter, OutNoStatFormatter
from cordex.writers.collocation_sentence_mapper import CollocationSentenceMapper


class Writer:
    @staticmethod
    def other_params(args):
        """ Obtains and formats parameters that are needed for initialization of Writer class. """
        return (args['multiple_output'], int(args['sort_by']), args['sort_reversed'], args['min_freq'])

    @staticmethod
    def make_output_writer(args, num_components, collocation_ids, word_renderer, is_ud):
        """ Returns an instance of Writer class with classical settings. """
        params = Writer.other_params(args)
        return Writer(args['out'], num_components, OutFormatter(collocation_ids, word_renderer, is_ud, args), args['collocation_sentence_map_dest'], params, args['separator'])

    @staticmethod
    def make_output_no_stat_writer(args, num_components, collocation_ids, word_renderer, is_ud):
        """ Returns an instance of Writer class with settings for no statistics output. """
        params = Writer.other_params(args)
        return Writer(args['out'], num_components, OutNoStatFormatter(collocation_ids, word_renderer, is_ud, args), args['collocation_sentence_map_dest'], params, args['separator'])

    def __init__(self, file_out, num_components, formatter, collocation_sentence_map_dest, params, separator):
        self.collocation_sentence_map_dest = collocation_sentence_map_dest
        if params is None:
            self.multiple_output = False
            self.sort_by = -1
            self.sort_order = None
            self.min_frequency = 1
        else:
            self.multiple_output = params[0]
            self.sort_by = params[1]
            self.sort_order = params[2]
            self.min_frequency = params[3]

        self.num_components = num_components
        self.output_file = file_out
        self.formatter = formatter
        self.separator = separator

    def header(self):
        """ Forms header content in a list with repeating columns first and one off columns after. """
        repeating_cols = self.formatter.header_repeat()
        cols = ["C{}_{}".format(i + 1, thd) for i in range(self.num_components) 
                for thd in repeating_cols]

        cols = ["Structure_ID"] + cols + ["Collocation_ID"]
        cols += self.formatter.header_right()
        return cols

    def sorted_rows(self, rows):
        """ Sorts rows by specified column. """
        if self.sort_by < 0 or len(rows) < 2:
            return rows

        if len(rows[0]) <= self.sort_by:
            logging.warning("Cannot sort by column #{}: Not enough columns!".format(len(rows[0])))
            return rows

        try:
            int(rows[0][self.sort_by])

            def key(row): 
                return int(row[self.sort_by])
        except ValueError:
            def key(row): 
                return row[self.sort_by].lower()

        return sorted(rows, key=key, reverse=self.sort_order)

    def write_header(self, file_handler, return_list):
        """ Writes header to output. """
        if return_list:
            return self.header()
        else:
            file_handler.write(self.separator.join(self.header()) + "\n")

    def write_out_worker(self, file_handler, structure, collocation_ids, col_sent_map, return_list):
        rows = []
        components = structure.components
        for match in collocation_ids.get_matches_for(structure):
            if len(match) < self.min_frequency:
                continue

            self.formatter.new_match(match)

            variable_word_order = self.find_variable_word_order(match.matches)

            if col_sent_map is not None:
                for words in match.matches:
                    token_ids = []
                    for int_i in range(1, len(words) + 1):
                        str_i = str(int_i)
                        if str_i in words:
                            token_ids.append(f'{words[str_i].sentence_id}.{words[str_i].id}')
                    col_sent_map.add_map(match.match_id, words['1'].sentence_id, '|'.join(token_ids))

            for words in match.matches:
                to_write = []

                idx = 1
                for _comp in components:
                    if _comp.idx == '#':
                        continue
                    idx_s = str(idx)
                    idx += 1
                    if idx_s not in words:
                        to_write.extend([""] * self.formatter.length())
                    else:
                        to_write.extend(self.formatter.content_repeat(words, match.representations, idx_s, structure.id))

                # make them equal size
                to_write.extend([""] * (self.num_components * self.formatter.length() - len(to_write)))

                # structure_id and collocation_id
                to_write = [structure.id] + to_write + [match.match_id]

                # header_right
                to_write.extend(self.formatter.content_right(len(match), variable_word_order))
                rows.append(to_write)
                break

        if rows != []:
            rows = self.sorted_rows(rows)
            if return_list:
                return [row for row in rows]
            else:
                file_handler.write("\n".join([self.separator.join(row) for row in rows]) + "\n")
                file_handler.flush()

        else:
            if return_list:
                return []

    def write_out(self, structures, collocation_ids, return_list=False):
        """ Writes processing results to file. """
        write_results = []

        if self.output_file is None and not return_list:
            return

        def fp_close(fp_):
            fp_.close()

        def fp_open(snum=None):
            if snum is None:
                return open(self.output_file, "w", encoding="UTF-8")
            else:
                return open("{}.csv".format(os.path.join(self.output_file, snum)), "w", encoding="UTF-8")

        # erase file/folder if exists
        if type(self.collocation_sentence_map_dest) == str and os.path.exists(self.collocation_sentence_map_dest):
            if os.path.isdir(self.collocation_sentence_map_dest):
                shutil.rmtree(self.collocation_sentence_map_dest)
            else:
                os.remove(self.collocation_sentence_map_dest)


        if not self.multiple_output:

            if return_list:
                write_results.append(self.header())
            else:
                fp = fp_open()
                self.write_header(fp, return_list)
            col_sent_map = CollocationSentenceMapper(self.collocation_sentence_map_dest) \
                if self.collocation_sentence_map_dest else None

        for s in progress(structures, "writing:{}".format(self.formatter)):
            if self.multiple_output:

                if return_list:
                    write_results.append(self.header())
                else:
                    fp = fp_open(s.id)
                    self.write_header(fp, return_list)
                if self.collocation_sentence_map_dest:
                    Path(self.collocation_sentence_map_dest).mkdir(parents=True, exist_ok=True)
                col_sent_map = CollocationSentenceMapper(os.path.join(self.collocation_sentence_map_dest, f'{s.id}.tsv')) \
                    if self.collocation_sentence_map_dest else None

            self.formatter.set_structure(s)
            if return_list:
                content = self.write_out_worker(None, s, collocation_ids, col_sent_map, return_list)
                write_results.extend(content)
            else:
                self.write_out_worker(fp, s, collocation_ids, col_sent_map, return_list)

                if self.multiple_output:
                    fp_close(fp)

        if not return_list:
            if not self.multiple_output:
                fp_close(fp)
            if col_sent_map:
                col_sent_map.close()

        if return_list and self.collocation_sentence_map_dest:
            if col_sent_map.get_mapper() is None:
                logging.warning("If you want to get collocation sentence map, you should set `collocation_sentence_map_dest` to True")
            return write_results, col_sent_map.get_mapper()

        if type(self.collocation_sentence_map_dest) == bool and self.collocation_sentence_map_dest:
            logging.error(
                "You should put path into `collocation_sentence_map_dest`. It can only be set to `True` if you are using `get_list()` function.")
        return write_results

    @staticmethod
    def find_variable_word_order(matches):
        """ Returns word order that has seen the most occurrences. """
        orders = {}
        for words in matches:
            order = tuple([tup[0] for tup in sorted(words.items(), key=lambda x: x[1].int_id)])
            orders[order] = orders.get(order, 0) + 1
        return max(orders, key=orders.get)



import re
from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
  long_description = f.read()

# read the version from classla/_version.py
version_file_contents = open(path.join(here, 'cordex/_version.py'), encoding='utf-8').read()
VERSION = re.compile('__cordex_version__ = \'(.*)\'').search(version_file_contents).group(1)

setup(name='cordex',
  version=VERSION,
  description=u"Parser for collocability",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url='https://github.com/clarinsi/cordex',
  author='CJVT',
  author_email='pypi@cjvt.si',
  license='MIT',
  packages=find_packages(),
  install_requires=[
    'conllu>=4.5.2',
    'cjvt-conversion-utils>=0.3',
    'importlib-resources>=5.4.0',
    'lxml>=4.9.1',
    'tqdm>=4.62.3',
    'zipp>=3.6.0',
    'requests>=2.21.0'
  ],
)


""" A file for handling mapper. """

class CollocationSentenceMapper:
    def __init__(self, output_dir):
        self.return_str = False
        if type(output_dir) == bool:
            self.output = []
            self.output.append(['Collocation_id', 'Sentence_id', 'Token_ids'])
            self.return_str = True
            return
        self.output = open(output_dir, "w")
        self.output.write(f'Collocation_id\tSentence_id\tToken_ids\n')

    def get_mapper(self):
        if self.return_str:
            return self.output

    def close(self):
        self.output.close()

    def add_map(self, collocation_id, sentence_id, token_ids):
        """ Adds a new line to output. """
        if self.return_str:
            self.output.append([collocation_id, sentence_id, token_ids])
        else:
            self.output.write(f'{collocation_id}\t{sentence_id}\t{token_ids}\n')



"""
Class for estimating remaining calculation time. Works only for multiple files input.
"""
from datetime import timedelta, datetime
import logging


class TimeInfo:
    def __init__(self, to_go):
        self.times = []
        self.to_go = to_go

    def add_measurement(self, time_measurement):
        """ Adds measurement to times for a better calculation. """
        if time_measurement > 0:
            self.times.append(time_measurement)
        self.to_go -= 1

    def info(self):
        """ Logs remaining time estimation. """
        seconds = sum(self.times) / len(self.times)
        td = timedelta(seconds=int(seconds * self.to_go))
        ft = datetime.now() + td
        logging.info("Going to finish in {}".format(ft.strftime("%d/%m @ %H:%M")))



"""
Word objects.
"""

from collections import defaultdict

from cordex.utils.converter import translate_msd


def prepare_ids(wid, fake_word):
    """ Gets sentence_id, word_id and int_word_id from a string id. """
    split_id = wid.split('.')
    word_id = split_id[-1]
    sentence_id = '.'.join(split_id[:-1]) if not fake_word else wid
    last_num = split_id[-1]
    if last_num[0] not in '0123456789':
        last_num = last_num[1:]
    int_word_id = int(last_num)
    return sentence_id, word_id, int_word_id


class WordDummy:
    """
    Fake word used in representation. These are used when none of the words in input satisfy checks (ie. no word with
    such msd). In these cases try to form word from inflectional lexicon.
    """
    def __init__(self, pos, lemma=None, text=None):
        self.xpos = pos
        self.udpos = pos
        self.lemma = lemma
        self.text = text


class Word:
    """
    Base word model.
    """
    def __init__(self, lemma, sentence_id, word_id, int_word_id, text, glue, fake_word=False):
        self.lemma = lemma

        self.idi = None
        self.text = text
        self.glue = glue
        self.previous_glue = False
        self.fake_word = fake_word

        self.links = defaultdict(list)

        self.id = word_id
        self.sentence_id = sentence_id
        self.int_id = int_word_id

    @staticmethod
    def from_tei_element(xml, do_msd_translate):
        """ Creates word from TEI word element. """
        raise NotImplementedError

    @staticmethod
    def from_conllu_element(token, sentence):
        """ Creates word from TEI word element. """
        raise NotImplementedError

    @staticmethod
    def get_msd(comp):
        """ Returns word msd. """
        raise NotImplementedError

    @staticmethod
    def pc_word(pc, do_msd_translate):
        """ Creates punctuation from TEI punctuation element. """
        raise NotImplementedError

    @staticmethod
    def fake_root_word(sentence_id):
        """ Creates a fake word. """
        raise NotImplementedError

    def add_link(self, link, to):
        """ Adds dependency parsing link to word. """
        self.links[link].append(to)

    def get_links(self, link):
        """ Returns links of specific type. """
        if link not in self.links and "|" in link:
            for l in link.split('|'):
                self.links[link].extend(self.links[l])

        return self.links[link]


class WordJOS(Word):
    def __init__(self, lemma, xpos, sentence_id, word_id, int_word_id, text, glue, do_msd_translate, fake_word=False):
        self.xpos = translate_msd(xpos, 'sl', lemma=lemma) if do_msd_translate else xpos

        super().__init__(lemma, sentence_id, word_id, int_word_id, text, glue, fake_word)

    @staticmethod
    def from_tei_element(xml, do_msd_translate):
        """ Creates word from TEI word element. """
        lemma = xml.get('lemma')
        xpos = WordJOS.get_msd(xml)
        wid = xml.get('id')
        text = xml.text
        glue = 'join' in xml.attrib and xml.get('join') == 'right'
        sentence_id, word_id, int_word_id = prepare_ids(wid, False)
        return WordJOS(lemma, xpos, sentence_id, word_id, int_word_id, text, glue, do_msd_translate)

    @staticmethod
    def from_conllu_element(token, sentence):
        """ Creates word from TEI word element. """
        glue = token['misc'] is not None and 'SpaceAfter' in token['misc'] and token['misc']['SpaceAfter'] == 'No'
        metadata = sentence.metadata['sent_id'] if 'sent_id' in sentence.metadata else ''
        return WordJOS(token['lemma'], token['xpos'], metadata, str(token['id']), int(token['id']), token['form'], glue, False)

    @staticmethod
    def get_msd(comp):
        """ Returns word msd. """
        d = dict(comp.items())
        assert 'ana' in d, 'Tag "ana" is not in element.'
        xpos = d['ana'][4:]

        return xpos

    @staticmethod
    def pc_word(pc, do_msd_translate):
        """ Creates punctuation from TEI punctuation element. """
        pc.set('lemma', pc.text)
        return WordJOS.from_tei_element(pc, do_msd_translate)

    @staticmethod
    def fake_root_word(sentence_id):
        """ Creates a fake word. """
        sentence_id, word_id, int_word_id = prepare_ids(sentence_id, True)
        return WordJOS('', '', sentence_id, word_id, int_word_id, '', False, False, True)


class WordUD(Word):
    def __init__(self, lemma, upos, sentence_id, word_id, int_word_id, text, glue, fake_word=False, feats=None):
        # udpos contains both upos and feats
        if feats:
            if upos:
                feats['POS'] = upos
                udpos = feats
            else:
                # goes here when POS already in feats
                udpos = feats

        else:
            if upos:
                udpos = {'POS': upos}
            else:
                udpos = ''

        self.udpos = udpos

        super().__init__(lemma, sentence_id, word_id, int_word_id, text, glue, fake_word)

    @staticmethod
    def from_tei_element(xml, do_msd_translate):
        """ Creates word from TEI word element. """
        lemma = xml.get('lemma')
        upos, feats = WordUD.get_msd(xml)
        wid = xml.get('id')
        text = xml.text
        glue = 'join' in xml.attrib and xml.get('join') == 'right'
        sentence_id, word_id, int_word_id = prepare_ids(wid, False)
        return WordUD(lemma, upos, sentence_id, word_id, int_word_id, text, glue, feats=feats)

    @staticmethod
    def from_conllu_element(token, sentence):
        glue = token['misc'] is not None and 'SpaceAfter' in token['misc'] and token['misc']['SpaceAfter'] == 'No'
        """ Creates word from TEI word element. """
        return WordUD(token['lemma'], token['upos'], sentence.metadata['sent_id'], str(token['id']), int(token['id']), token['form'], glue, feats=token['feats'])

    @staticmethod
    def get_msd(comp):
        """ Returns word msd. """
        d = dict(comp.items())
        assert 'msd' in d, 'Tag "msd" is not in element.'
        feats_expanded = {attrib.split('=')[0]: attrib.split('=')[1] for attrib in d['msd'].split('|')}
        if 'UPosTag' in feats_expanded:
            upos = feats_expanded['UPosTag']
            del feats_expanded['UPosTag']
        else:
            upos = feats_expanded['UposTag']
            del feats_expanded['UposTag']
        feats = feats_expanded
        return upos, feats

    @staticmethod
    def pc_word(pc, do_msd_translate):
        """ Creates punctuation from TEI punctuation element. """
        pc.set('lemma', pc.text)
        return WordUD.from_tei_element(pc, do_msd_translate)

    @staticmethod
    def fake_root_word(sentence_id):
        """ Creates a fake word. """
        sentence_id, word_id, int_word_id = prepare_ids(sentence_id, True)
        return WordUD('', '', sentence_id, word_id, int_word_id, '', False, True)



"""
Classes related to syntactic structure components.
"""
from enum import Enum

from cordex.structures.order import Order
from cordex.representations.representation_assigner import RepresentationAssigner
from cordex.restrictions.restriction_group import RestrictionGroup


class ComponentStatus(Enum):
    Optional = 0
    Required = 1
    Forbidden = 2


class ComponentType(Enum):
    Other = 0
    Core = 2
    Core2w = 3


class Component:
    def __init__(self, info, is_ud):
        idx = info['cid']
        name = info['label'] if 'label' in info else None
        typ = ComponentType.Core if info['type'] == "core" else ComponentType.Other

        if 'status' not in info:
            status = ComponentStatus.Required
        elif info['status'] == 'forbidden':
            status = ComponentStatus.Forbidden
        elif info['status'] == 'obligatory':
            status = ComponentStatus.Required
        elif info['status'] == 'optional':
            status = ComponentStatus.Optional
        else:
            raise NotImplementedError("strange status: {}".format(info['status']))

        self.is_ud = is_ud
        self.status = status
        self.name = name
        self.idx = idx
        self.restrictions = RestrictionGroup([None], self.is_ud) if 'restriction' in info else []
        self.children = []
        self.representation = []
        self.selection = {}
        self.type = typ

        self.iter_ctr = 0

    def add_child(self, child_component, link_label, order):
        """ Adds next component"""
        self.children.append((child_component, link_label, Order.new(order)))

    def set_restriction(self, restrictions_tags):
        """ Set regex restrictions to component. """
        if not restrictions_tags:
            self.restrictions = RestrictionGroup([None], self.is_ud)

        # if first element is of type restriction all following are as well
        elif restrictions_tags[0].tag == "restriction":
            self.restrictions = RestrictionGroup(restrictions_tags, self.is_ud)

        # combinations of 'and' and 'or' restrictions are currently not implemented
        elif restrictions_tags[0].tag == "restriction_or":
            self.restrictions = RestrictionGroup(restrictions_tags[0], self.is_ud, group_type='or')

        else:
            raise RuntimeError("Unreachable")

    def set_representation(self, representation):
        """ Set component representation. """
        for rep in representation:
            crend = RepresentationAssigner()
            for feature in rep:
                crend.add_feature(feature.attrib)
            self.representation.append(crend)

    def create_children(self, deps, comps, restrs, reprs):
        """ Create component children. """
        to_ret = []
        for d in deps:
            if d[0] == self.idx:
                _, idx, dep_label, order = d

                child = Component(comps[idx], self.is_ud)
                child.set_restriction(restrs[idx])
                child.set_representation(reprs[idx])
                to_ret.append(child)

                self.add_child(child, dep_label, order)
                others = child.create_children(deps, comps, restrs, reprs)
                to_ret.extend(others)

        return to_ret

    def match(self, word):
        """ Returns list of words that match restrictions. When no words are found return None. """
        # check self restrictions
        m1 = self._match_self(word)
        if m1 is None:
            return None

        # check next node
        mn = self._match_next(word)
        if mn is None:
            return None

        to_ret = [m1]
        for cmatch in mn:
            # if good match but nothing to add, just continue
            if len(cmatch) == 0:
                continue

            # create new to_ret, to which extend all results
            new_to_ret = []
            for tr in to_ret:
                # make sure that one word is not used twice in same to_ret
                new_to_ret.extend([{**dict(tr), **m} for m in cmatch if all([m_v not in dict(tr).values() for m_v in m.values()])])
            if len(new_to_ret) == 0:
                return None
            to_ret = new_to_ret
            del new_to_ret

        return to_ret

    def _match_self(self, word):
        """ Checks restrictions of root word. """
        # matching
        if self.restrictions.match(word):
            return {self.idx: word}

    def _match_next(self, word):
        """ Matches for every component in links from this component """
        to_ret = []

        # need to get all links that match
        for next, link, order in self.children:
            next_links = word.get_links(link)
            to_ret.append([])

            # good flag
            good = next.status != ComponentStatus.Required
            for next_word in next_links:
                if not order.match(word, next_word):
                    continue

                match = next.match(next_word)

                if match is not None:
                    # special treatment for forbidden
                    if next.status == ComponentStatus.Forbidden:
                        good = False
                        break

                    else:
                        assert type(match) is list
                        to_ret[-1].extend(match)
                        good = True

            # if none matched, nothing found!
            if not good:
                return None

        return to_ret
